require 'spec_helper'

describe RubyAwaitNode do
  def return_42
    "(function() { var f = 42; return 42;})()"
  end

  def await_return_42
    "(async function() { var f = 42; await new Promise(resolve => setTimeout(resolve, 1000)); return 42;})()"
  end

  def create_ruby_await_node_runtime
    RubyAwaitNode::Runtime.new(:debug => false)
  end

  def create_ruby_await_node_context
    RubyAwaitNode::Context.new(create_ruby_await_node_runtime)
  end

  context "simple javascript" do
    it "returns desired value" do
      js_result = create_ruby_await_node_context.eval(return_42)
      expect(js_result).to eq 42
    end
  end

  context "await javascript" do
    it "return desired value" do
      js_result = create_ruby_await_node_context.eval(await_return_42)
      expect(js_result).to eq 42
    end
  end

  describe "load an entry file" do
    it "loads correctly" do
      path = File.expand_path("sample_entry.js", __dir__)
      context = create_ruby_await_node_context
      context.load(path)
      js_result = context.eval("global.entryName")
      expect(js_result).to eq "sample"
    end
  end

  it "raises an RubyAwaitNode::ProgramError on error" do
    expect {
      create_ruby_await_node_context.eval("(function() { throw new Error('foo\\nbar', 0, 'test.js'); })()")
    }.to raise_error(RubyAwaitNode::ProgramError)
  end

  it "requires js to be in a self-executing function" do
    js_result = -1

    expect {
      js_result = create_ruby_await_node_context.eval("return true;")
    }.to raise_error(RubyAwaitNode::RuntimeError)

    expect(js_result).to eq(-1)
  end

  it "safely allows subsequent #eval calls after runtime error" do
    js_result = create_ruby_await_node_context.eval(return_42)
    expect(js_result).to eq 42

    expect {
      js_result = create_ruby_await_node_context.eval("return true;")
    }.to raise_error(RubyAwaitNode::RuntimeError)

    js_result = create_ruby_await_node_context.eval(return_42)
    expect(js_result).to eq 42
  end

  it "should be thread safe when sharing context between threads" do
    thread_count = 128
    threads = []
    semaphore = Mutex.new
    results = []

    context = create_ruby_await_node_context
    context.eval("global.a = 1.0;")

    thread_count.times { |index|
      threads << Thread.new {
        result = context.call("(function(b) { return (global.a * b); })", index)
        semaphore.synchronize {
          results << result
        }
      }
    }

    threads.each { |t| t.join }

    sum = results.inject(0) { |result, element| result + element }

    expect(sum).to eq(8128) # http://www.wolframalpha.com/input/?i=127th+trianglur+number
  end

  it "should be thread safe when creating new contexts in threads" do
    thread_count = 128
    threads = []
    semaphore = Mutex.new
    results = []

    thread_count.times { |index|
      threads << Thread.new {
        context = create_ruby_await_node_context
        context.eval("global.a = 2.0;") # this asserts that the context is shared
        result = context.call("(function(b) { return (a * b); })", index)
        semaphore.synchronize {
          results << result
        }
      }
    }

    threads.each { |t| t.join }

    sum = results.inject(0) { |result, element| result + element }

    expect(sum).to eq(8128 * 2) # http://www.wolframalpha.com/input/?i=127th+trianglur+number
  end

  it "should not break if exception occurs in thread" do
    context_a = create_ruby_await_node_context

    g_err_a = nil
    g_err_b = nil

    make_program_error_thread = Thread.new {
      begin
        _b = context_a.call("(function() { asd() })")
      rescue => err_a
        g_err_a = err_a
      end
    }

    context_b = create_ruby_await_node_context

    begin
      _c = context_b.call("(function() { asd) })")
    rescue => err_b
      g_err_b = err_b
    end

    make_program_error_thread.join

    expect(g_err_a).to be_kind_of(RubyAwaitNode::ProgramError)
    expect(g_err_b).to be_kind_of(RubyAwaitNode::RuntimeError)
  end

  it "should work when the context is shared between forked processes" do
    fork_test_harness_pid = Process.spawn("bundle exec ruby spec/fork_test_harness.rb")
    exit_pid, exit_status = Process.wait2(fork_test_harness_pid)

    expect(exit_pid).to eq(fork_test_harness_pid)
    expect(exit_status).to eq(0)
  end
end
